/*
# Copyright (c) 2003-2004 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
# ----------------------------------------------------------------------------
#
# Primary author: Purushotham Kamath
#                 July 2003
#This material is based upon work supported by the Defense Advanced Research 
#Projects Agency under contract no. N66001-02-1-8939 issued by the Space and 
#Naval Warfare Systems Center (SPAWAR).  Any opinions, findings, and conclusions 
#or recommendations expressed in this material are those of the author(s) and 
#do not necessarily reflect the views of the Defense Advanced Research Projects
#Agency, SPAWAR, or the U.S. Government.
*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include <packet.h>
#include <eventhandler.h>
#include <queue.h>
#include <scheduler.h>
#include <ipport.h>
#include <sense.h>
#include <line.h>
#include <debug.h>


extern Scheduler * sched;
extern Sense *s;
extern Line *l;
extern IPPort *ip;
extern char inputfile[20];
extern char codefile[20];
extern double load;


IPPort::IPPort( int s):EventHandler()
{
	fin = fopen(inputfile, "r");
	if (fin == NULL)
		dprintf("res.cc: IPPort::IPPort(): Could not open input file for reading\n");
	//Inputport delay: No delays on the input port, packets are scheduled as per their timestamp
	spreading = s;
	InitializeCodeSet();
}

IPPort::~IPPort()
{
	fclose(fin);
}


//The data acquisition function of the IPPort: This function reads a 
//trace file, constructs a packet and processes it.
void IPPort::read()
{
	dprintf("res.cc: IPPort::read(): Entered function\n");
	

	double timestamp;
	int pktid;
	int srcaddr;
	int srcport;
	int dstaddr;
	int dstport;
	int size;

//NS format
//Read event from trace file
	if (fscanf(fin,  "%lf %d %d.%d %d.%d %d\n", &timestamp, &size, &srcaddr, &srcport, &dstaddr, &dstport, &pktid) == EOF)
	{
		//sched->halt = 1;
		return;
	}
		
//Call ipport's processing function.
	dprintf("res.cc: IPPort::read(): Packetid: %d Dst addr: %d Size: %d\n", pktid, dstaddr, size);
	dprintf("res.cc: IPPort::read(): Current time: %.12f\n", sched->currenttime);
	dprintf("res.cc: IPPort::read(): Timestamp: %.12f\n", timestamp);

	dprintf("Got a packet: Timestamp: %0.9f\n", timestamp);
	Packet *packet = new Packet();
	packet->id = pktid;
	packet->opport = 0; //Default route 
	packet->altport = 0;
	packet->srcaddr = srcaddr;
	packet->dstaddr = dstaddr;
	packet->srcport = srcport;
	packet->dstport = dstport;
	packet->size = size;
	packet->timestamp = timestamp;
	strcpy(packet->tag, "E"); //The default tag is "Electrical"
	
	int rcvd = (int)(floor((random()/(pow(2, 31) -1)) * (numcodes))); 
	packet->codeword = code[rcvd];


	double delay  = timestamp - sched->currenttime;
	dprintf("ipport::read: Scheduling for %.6f\n", delay);

	sched->scheduleevent(this, packet, delay );

	static int numpkts =0;
	numpkts++;
}

//The event handling function of the IPPort. It dequeues a packet and
//sends it to the next component.
void IPPort::handleevent(Packet *p)
{
	dprintf("res.cc: IPPort::handleevent(): Entered function\n");
	

	//Pass it on to the next component in the path.
	dprintf("res.cc: IPPort::handleevent(): Packet id: %d\n", p->id);
	dprintf("res.cc: IPPort::handleevent(): Packet processing complete, passing it on to the next component.\n");

	s->process(p);
	read();
}

void IPPort::dumpstats()
{
	printf("res.cc: IPPort::dumpstats():\n");
}

void IPPort::InitializeCodeSet()
{

	FILE *cf = fopen(codefile, "r");
	if (cf == NULL)
	{
		printf("Interference: Error opening file: %s\n", codefile);
		exit(-1);
	}
	int error = fscanf(cf, "%d", &numcodes);
	if(error < 0)
	{
		printf("Interference: Error reading file\n");
		exit(-1);
	}

	error = fscanf(cf, "%d", &weight);
	if(error < 0)
	{
		printf("Interference: Error reading file\n");
		exit(-1);
	}
	dprintf("Num code: %d\n", numcodes);
	dprintf("Weight: %d\n", weight);

	code = new unsigned int [numcodes];
	for(int count = 0; count < numcodes; count++)
	{
		fscanf(cf, "%d", &code[count]);
		dprintf("%d\n", code[count]);
	}
	
	fclose(cf);

}


int IPPort:: rightshift( int code, int shift)
{
    int shifted =code;
	for(int count=0; count< shift; count++)
	{
		int lsb = shifted & 1;
//      printf("lsb: %d\n", lsb);
		if (lsb != 1)
			shifted = shifted >> 1;
		else
			shifted = (shifted >> 1) | (int)(pow(2, spreading - 1)) ;
	}
	return shifted;
}

int IPPort:: leftshift( int code, int shift)
{
    int shifted =code;
	for(int count=0; count< shift; count++)
	{
		int msb = shifted & (int)pow(2, spreading -1);
//      printf("lsb: %d\n", lsb);
		if (msb ==0)
			shifted = shifted << 1;
		else
			shifted = (shifted << 1) | 1 ;

		shifted = shifted & ((int)pow(2, spreading) - 1);
        dprintf("shifted: %d\n", shifted);
	}
	return shifted;
}
